package sim.field.grid;
import sim.util.*;

/**
   A wrapper for 2D arrays of ints.

   <p>This object expects that the 2D arrays are rectangular.  You are encouraged to access the array
   directly.  The object
   implements all of the Grid2D interface.  See Grid2D for rules on how to properly implement toroidal
   or hexagonal grids.
    
   <p>The width and height of the object are provided to avoid having to say field[x].length, etc.  
*/

public /*strictfp*/ class IntGrid2D extends AbstractGrid2D
    {
    public int[/**x*/][/**y*/] field;
    
    public IntGrid2D (int xdim, int ydim)
        {
        width = xdim;
        height = ydim;
        field = new int[xdim][ydim];
        }
    
    public IntGrid2D (int xdim, int ydim, int initialValue)
        {
        this(xdim,ydim);
        setTo(initialValue);
        }
    
    public IntGrid2D (IntGrid2D values)
        {
        setTo(values);
        }

    public final void set(final int x, final int y, final int val)
        {
        field[x][y] = val;
        }
    
    public final int get(final int x, final int y)
        {
        return field[x][y];
        }

    public final IntGrid2D setTo(int thisMuch)
        {
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                fieldx[y]=thisMuch;
	    }
        return this;
        }

    public final IntGrid2D setTo(IntGrid2D values)
        {
        if (width != values.width || height != values.height)
            {
	    final int width = this.width = values.width;
	    final int height = this.height = values.height;
            field = new int[width][];
            for(int x =0 ; x < width; x++)
                field[x] = (int []) (values.field[x].clone());
            }
        else
            {
            for(int x =0 ; x < width; x++)
                System.arraycopy(values.field[x],0,field[x],0,height);
            }

        return this;
        }

    public final int max()
        {
        int max = Integer.MIN_VALUE;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                if (max < fieldx[y]) max = fieldx[y];
	    }
        return max;
        }

    public final int min()
        {
        int min = Integer.MAX_VALUE;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                if (min > fieldx[y]) min = fieldx[y];
	    }
        return min;
        }
        
    public final double mean()
        {
        long count = 0;
        double mean = 0;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                { mean += fieldx[y]; count++; }
	    }
        return (count == 0 ? 0 : mean / count);
        }
        
    public final IntGrid2D upperBound(int toNoMoreThanThisMuch)
        {
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                if (fieldx[y] > toNoMoreThanThisMuch)
                    fieldx[y] = toNoMoreThanThisMuch;
	    }
        return this;
        }

    public final IntGrid2D lowerBound(int toNoLowerThanThisMuch)
        {
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                if (fieldx[y] < toNoLowerThanThisMuch)
                    fieldx[y] = toNoLowerThanThisMuch;
	    }
        return this;
        }

    public final IntGrid2D add(int withThisMuch)
        {
        if (withThisMuch==0.0) return this;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                fieldx[y]+=withThisMuch;
	    }
        return this;
        }
        
    public final IntGrid2D add(IntGrid2D withThis)
        {
        int[][]ofield = withThis.field;
        int[] ofieldx = null;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
	    ofieldx = ofield[x];
            for(int y=0;y<height;y++)
                fieldx[y]+=ofieldx[y];
	    }
        return this;
        }

    public final IntGrid2D multiply(int byThisMuch)
        {
        if (byThisMuch==1.0) return this;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
            for(int y=0;y<height;y++)
                fieldx[y]*=byThisMuch;
	    }
        return this;
        }
    
    public final IntGrid2D multiply(IntGrid2D withThis)
        {
        int[][]ofield = withThis.field;
        int[] ofieldx = null;
	int[] fieldx = null;
	final int width = this.width;
	final int height = this.height;
        for(int x=0;x<width;x++)
	    {
	    fieldx = field[x];
	    ofieldx = ofield[x];
            for(int y=0;y<height;y++)
                fieldx[y]*=ofieldx[y];
	    }
        return this;
        }


    /*
     * Gets all neighbors of a location that satisfy max( abs(x-X) , abs(y-Y) ) <= d
     * Returns the neighbors and their x and y positions; xPos and yPos can be null if you don't care about the x and y positions.
     */
    public final void getNeighborsMaxDistance( final int x, final int y, final int dist, final boolean toroidal, IntBag result, IntBag xPos, IntBag yPos )
	{
        if( xPos == null )
            xPos = new IntBag();
        if( yPos == null )
            yPos = new IntBag();

        getNeighborsMaxDistance( x, y, dist, toroidal, xPos, yPos );

        result.clear();
        for( int i = 0 ; i < xPos.numObjs ; i++ )
            result.add( field[xPos.objs[i]][yPos.objs[i]] );
	}

    /*
     * Gets all neighbors of a location that satisfy abs(x-X) + abs(y-Y) <= d
     * Returns the neighbors and their x and y positions; xPos and yPos can be null if you don't care about the x and y positions.
     */
    public final void getNeighborsHamiltonianDistance( final int x, final int y, final int dist, final boolean toroidal, IntBag result, IntBag xPos, IntBag yPos )
	{
        if( xPos == null )
            xPos = new IntBag();
        if( yPos == null )
            yPos = new IntBag();

        getNeighborsHamiltonianDistance( x, y, dist, toroidal, xPos, yPos );

        result.clear();
        for( int i = 0 ; i < xPos.numObjs ; i++ )
            result.add( field[xPos.objs[i]][yPos.objs[i]] );
	}

    /*
     * Gets all neighbors of a location in hexagonal world
     * Returns the neighbors and their x and y positions; xPos and yPos can be null if you don't care about the x and y positions.
     */
    public final void getNeighborsHexagonalDistance( final int x, final int y, final int dist, final boolean toroidal, IntBag result, IntBag xPos, IntBag yPos )
	{
        if( xPos == null )
            xPos = new IntBag();
        if( yPos == null )
            yPos = new IntBag();

        getNeighborsHexagonalDistance( x, y, dist, toroidal, xPos, yPos );

        result.clear();
        for( int i = 0 ; i < xPos.numObjs ; i++ )
            result.add( field[xPos.objs[i]][yPos.objs[i]] );
	}

    }
